@@ -6,7 +6,7 @@
 ;; Maintainer: Protesilaos Stavrou General Issues <~protesilaos/general-issues@lists.sr.ht>
 ;; URL: https://git.sr.ht/~protesilaos/beframe
 ;; Mailing-List: https://lists.sr.ht/~protesilaos/general-issues
-;; Version: 0.1.11
+;; Version: 0.2.0
 ;; Package-Requires: ((emacs "28.1"))
 
 ;; This file is NOT part of GNU Emacs.
@@ -29,101 +29,14 @@
 ;; `beframe' enables a frame-oriented Emacs workflow where each frame
 ;; has access to the list of buffers visited therein.  In the interest
 ;; of brevity, we call buffers that belong to frames "beframed".
-;; Beframing is achieved in three main ways:
 ;;
-;; 1. By calling the command `beframe-switch-buffer'.  It is like the
-;;    standard `switch-to-buffer' except the list of candidates is
-;;    limited to those that the current frame knows about.
-;;
-;; 2. By enabling the global minor mode `beframe-mode'.  It sets the
-;;    `read-buffer-function' to one that filters buffers per frame.  As
-;;    such, commands like `switch-to-buffer', `next-buffer', and
-;;    `previous-buffer' automatically work in a beframed way.
-;;
-;; 3. The command `beframe-buffer-menu' produces a dedicated buffer with
-;;    a list of buffers for the current frame.  This is the counterpart
-;;    of `beframe-switch-buffer'.  When called with a prefix argument
-;;    (`C-u' with default key bindings), it prompts for a frame whose
-;;    buffers it will display.
-;;
-;; Producing multiple frames does not generate multiple buffer lists.
-;; There still is only one global list of buffers.  Beframing them simply
-;; filters the list.
-;;
-;; The user option `beframe-global-buffers' contains a list of strings
-;; that represent buffers names.  Those buffers are never beframed and
-;; are available in all frames.  The default value contains the buffers
-;; `*scratch*', `*Messages*', and `*Backtrace*'.
-;;
-;; The user option `beframe-create-frame-scratch-buffer' allows
-;; `beframe-mode' to create a frame-specific scratch buffer that runs
-;; the `initial-major-mode'.  This is done upon the creation of a new
-;; frame and the scratch buffer is named after the frame it belongs
-;; to.  For example, if the frame is called `modus-themes', the
-;; corresponding scratch buffer is `*scratch for modus-themes*'.  Set
-;; this user option to `nil' to disable the creation of such scratch
-;; buffers.
-;;
-;; The user option `beframe-kill-frame-scratch-buffer' is the
-;; counterpart of `beframe-create-frame-scratch-buffer'.  It kills the
-;; frame-specific scratch buffer after the frame is deleted.  Set this
-;; user option to `nil' to disable the killing of such buffers.
-;;
-;; The command `beframe-assume-frame-buffers' (alias
-;; `beframe-add-frame-buffers') prompts for a frame and then copies
-;; its buffer list into the current frame.
-;;
-;; The command `beframe-unassume-frame-buffers' (alias
-;; `beframe-remove-frame-buffers') prompts for a frame and then
-;; removes its buffer list from the current frame.
-;;
-;; The `beframe-mode' does the following:
-;;
-;; - Sets the value of `read-buffer-function' to a function that
-;;   beframes all commands that read that variable.  This includes the
-;;   likes of `switch-to-buffer', `next-buffer', and `previous-buffer'.
-;;
-;; - Add a filter to newly created frames so that their
-;;   `buffer-predicate' parameter beframes buffers.
-;;
-;; - Renames newly created frames so that they have a potentially more
-;;   meaningful title.  The user option `beframe-rename-function'
-;;   specifies the function that handles this process.  When its value
-;;   is nil, no renaming is performed.
-;;
-;; - When the user option `beframe-functions-in-frames' contains a list
-;;   of functions, it makes them run with `other-frame-prefix', meaning
-;;   that they are called in a new frame.  The default value of that
-;;   user option affects the `project.el' project-switching selection:
-;;   the new project buffer appears in its own frame and, thus, becomes
-;;   part of a beframed list of buffers, isolated from all other frames.
-;;
-;; - Handles the creation and deletion of frame-specific scratch
-;;   buffers,per the user options `beframe-create-frame-scratch-buffer',
-;;   `beframe-kill-frame-scratch-buffer'.
-;;
-;; Development note: `beframe' is in its early days.  The minor mode may
-;; be revised to have more features and/or greater flexibility.
-;;
-;; The `consult' package by Daniel Mendler (also known as @minad)
-;; provides commands that extend the functionality of the generic
-;; minibuffer.  Among them is `consult-buffer'.  With that command the
-;; user can search for buffers, recent files, and bookmarks, among
-;; others.
-;;
-;; `consult-buffer' displays a global list of buffers and is never
-;; beframed.  This is done by design.  Instead of forcing it to only show
-;; beframed buffers, we get the best of both worlds: (i) a convenient way
-;; to access the global list of buffers and (ii) a new entry to the
-;; `consult-buffer-sources' that registers beframed buffers as their own
-;; group.  Users can narrow directly to this group by typing `F' followed
-;; by `SPC'.
-;;
-;; Backronym: Buffers Encapsulated in Frames Realise Advanced
-;; Management of Emacs.
+;; Read the official manual for detailed documentation:
+;; <https://protesilaos.com/emacs/beframe>.
 
 ;;; Code:
 
+(eval-when-compile (require 'cl-lib))
+
 (defgroup beframe ()
   "Isolate buffers per frame."
   :group 'frames)
@@ -205,24 +118,59 @@ If nil, no renaming is performed."
      (get-buffer name))
    beframe-global-buffers))
 
-(defun beframe--buffer-list (&optional frame)
+(cl-defun beframe-buffer-list (&optional frame &key sort)
   "Return list of buffers that are used by the current frame.
 With optional FRAME as an object that satisfies `framep', return
 the list of buffers that are used by FRAME.
 
-Include `beframe-global-buffers' in the list."
-  (delq nil
-        (delete-dups
-         (append (beframe--frame-buffers frame)
-                 (beframe--global-buffers)))))
+The key SORT may be a function taking the list of buffers as an
+argument, and returning a new list to be used instead.  This can,
+for example, be used to prefer hidden buffers to visible ones—see
+`beframe-buffer-sort-visibility'.
 
-(defun beframe--buffer-names (&optional frame)
-  "Return list of names of `beframe--buffer-list' as strings.
-With optional FRAME, do it for the given frame name."
+Include `beframe-global-buffers' in the list."
+  (funcall (or sort #'identity)
+           (delq nil
+                 (delete-dups
+                  (append (beframe--frame-buffers frame)
+                          (beframe--global-buffers))))))
+
+(defun beframe--buffer-list-consolidated ()
+  "Return list of buffers from all frames.
+This is the same as the output of the `buffer-list' function
+minus all the internal buffers."
+  (seq-filter
+   (lambda (buf)
+     (and (bufferp buf)
+          (not (string-prefix-p " " (buffer-name buf)))))
+   (buffer-list)))
+
+(define-obsolete-function-alias
+  'beframe--buffer-list
+  'beframe-buffer-list
+  "0.2.0")
+
+(cl-defun beframe-buffer-names (&optional frame &key sort)
+  "Return list of names of `beframe-buffer-list' as strings.
+With optional FRAME, do it for the given frame name.  With key
+SORT, apply this sorting function—see `beframe-buffer-list' for
+more information."
   (mapcar
    (lambda (buf)
      (buffer-name buf))
-   (beframe--buffer-list frame)))
+   (beframe-buffer-list frame :sort sort)))
+
+(defun beframe--buffer-names-consolidated ()
+  "Return list of names of all buffers as strings."
+  (mapcar
+   (lambda (buf)
+     (buffer-name buf))
+   (beframe--buffer-list-consolidated)))
+
+(define-obsolete-function-alias
+  'beframe--buffer-names
+  'beframe-buffer-names
+  "0.2.0")
 
 (defun beframe--read-buffer-p (buf &optional frame)
   "Return non-nil if BUF belongs to the current FRAME.
@@ -233,7 +181,7 @@ it must satisfy `framep'."
     (when (consp buf)
       (setq b (car buf)))
     (unless (string-prefix-p " " b)
-      (seq-contains-p (beframe--buffer-names frame) b))))
+      (seq-contains-p (beframe-buffer-names frame) b))))
 
 (defvar beframe-history nil
   "Minibuffer history of frame specific buffers.")
@@ -245,7 +193,7 @@ PROMPT, DEF, REQUIRE-MATCH, and PREDICATE are the same as
 `read-buffer'.  The PREDICATE is ignored, however, to apply the
 per-frame filter."
   (completing-read prompt
-                   (beframe--buffer-names)
+                   (beframe-buffer-names)
                    #'beframe--read-buffer-p
                    require-match
                    nil
@@ -253,7 +201,7 @@ per-frame filter."
                    def))
 
 (defun beframe--buffer-prompt (&optional frame)
-  "Prompt for buffer among `beframe--buffer-names'.
+  "Prompt for buffer among `beframe-buffer-names'.
 
 Use the previous buffer as the default value, such that
 subsequent invocations of this command flip between the current
@@ -338,16 +286,18 @@ The window manager must permit such an operation.  See bug#61319:
   (select-frame-set-input-focus frame)
   (switch-to-buffer buffer))
 
-(defun beframe--list-buffers-noselect (&optional frame)
+(cl-defun beframe-list-buffers-noselect (&optional frame &key sort)
   "Produce a buffer list of buffers for optional FRAME.
-When FRAME is nil, use the current one.
+When FRAME is nil, use the current one.  With key SORT, apply
+this sorting function—see `beframe-buffer-list' for more
+information.
 
 This is a simplified variant of `list-buffers-noselect'."
   (let* ((frame (if (framep frame) frame (selected-frame)))
          (name (frame-parameter frame 'name))
          (old-buf (current-buffer))
          (buf (get-buffer-create (format "*Buffer List for %s*" name)))
-         (buffer-list (beframe--buffer-list frame)))
+         (buffer-list (beframe-buffer-list frame :sort sort)))
     (with-current-buffer buf
       (Buffer-menu-mode)
       (setq-local Buffer-menu-files-only nil
@@ -357,11 +307,17 @@ This is a simplified variant of `list-buffers-noselect'."
       (tabulated-list-print))
     buf))
 
+(define-obsolete-function-alias
+  'beframe--list-buffers-noselect
+  'beframe-list-buffers-noselect
+  "0.2.0")
+
 ;;;###autoload
-(defun beframe-buffer-menu (&optional frame)
+(cl-defun beframe-buffer-menu (&optional frame &key sort)
   "Produce a `buffer-menu' for the current FRAME.
 With FRAME as a prefix argument, prompt for a frame.  When called
-from Lisp, FRAME satisfies `framep'.
+from Lisp, FRAME satisfies `framep'.  With key SORT, apply this
+sorting function—see `beframe-buffer-list' for more information.
 
 The bespoke buffer menu is displayed in a window using
 `display-buffer'.  Configure `display-buffer-alist' to control
@@ -370,7 +326,33 @@ its placement and other parameters."
    (list
     (when current-prefix-arg
       (beframe--frame-object (beframe--frame-prompt)))))
-  (display-buffer (beframe--list-buffers-noselect frame)))
+  (display-buffer (beframe-list-buffers-noselect frame :sort sort)))
+
+(defun beframe--assume (buffers)
+  "Unassume BUFFERS from current frame.
+BUFFERS is a list of buffer objects.  If BUFFERS satisfies
+`framep', then the list of buffers is that of the corresponding
+frame object (per `beframe-buffer-list')."
+  (let ((buffer-list
+         (delete-dups
+          (append
+           (if (framep buffers)
+               (beframe-buffer-list buffers)
+             buffers)
+           (beframe-buffer-list)))))
+    (modify-frame-parameters nil `((buffer-list . ,buffer-list)))))
+
+(defun beframe--unassume (buffers)
+  "Unassume BUFFERS from current frame.
+BUFFERS is a list of buffer objects.  If BUFFERS satisfies
+`framep', then the list of buffers is that of the corresponding
+frame object (per `beframe-buffer-list')."
+  (let ((buffer-list
+         (seq-filter
+          (lambda (buf)
+            (not (member buf (if (framep buffers) (beframe-buffer-list buffers) buffers))))
+          (beframe-buffer-list))))
+    (modify-frame-parameters nil `((buffer-list . ,buffer-list)))))
 
 ;;;###autoload
 (defun beframe-assume-frame-buffers (frame)
@@ -378,13 +360,10 @@ its placement and other parameters."
 When called interactively, prompt for FRAME using completion.
 Else FRAME must satisfy `framep'.
 
-Also see `beframe-unassume-frame-buffers'."
+Also see `beframe-unassume-frame-buffers',
+`beframe-assume-buffers', `beframe-unassume-buffers'."
   (interactive (list (beframe--frame-object (beframe--frame-prompt))))
-  (let* ((other-buffer-list (beframe--buffer-list frame))
-         (buffers (delete-dups (append other-buffer-list (beframe--buffer-list)))))
-    (modify-frame-parameters
-     nil
-     `((buffer-list . ,buffers)))))
+  (beframe--assume frame))
 
 (defalias 'beframe-add-frame-buffers 'beframe-assume-frame-buffers
   "Alias of `beframe-assume-frame-buffers' command.")
@@ -395,23 +374,105 @@ Also see `beframe-unassume-frame-buffers'."
 When called interactively, prompt for FRAME using completion.
 Else FRAME must satisfy `framep'.
 
-Also see `beframe-assume-frame-buffers'."
+Also see `beframe-assume-frame-buffers',
+`beframe-assume-buffers', `beframe-unassume-buffers'."
   (interactive (list (beframe--frame-object (beframe--frame-prompt))))
-  (let* ((other-buffer-list (beframe--buffer-list frame))
-         (new-buffer-list
-          (seq-filter
-           (lambda (buf)
-             (not (member buf other-buffer-list)))
-           (beframe--buffer-list))))
-    (modify-frame-parameters
-     nil
-     `((buffer-list . ,new-buffer-list)))))
+  (beframe--unassume frame))
 
 (defalias 'beframe-remove-frame-buffers 'beframe-unassume-frame-buffers
   "Alias of `beframe-unassume-frame-buffers' command.")
 
-;; TODO 2023-03-02: Define commands to `completing-read-multiple'
-;; assume/unassume buffers.
+(defun beframe--buffers-name-to-objects (buffers)
+  "Convert list of named BUFFERS to their corresponding objects."
+  (mapcar
+   (lambda (buf)
+     (get-buffer buf))
+   buffers))
+
+(defun beframe--buffer-list-prompt-crm (&optional frame)
+  "Select one or more buffers in FRAME separated by `crm-separator'.
+Optional FRAME argument is an object that satisfies `framep'.  If
+FRAME is nil, the current frame is used.  If FRAME is non-nil but
+not a frame object, treat it as a flag for the consolidated
+buffer list (buffers from all frames)."
+  (completing-read-multiple
+   "Select buffers: "
+   (cond
+    ((framep frame)
+     (beframe-buffer-names frame))
+    (frame
+     (beframe--buffer-names-consolidated))
+    (t (beframe-buffer-names)))
+   nil
+   :require-match))
+
+;;;###autoload
+(defun beframe-assume-buffers (buffers)
+  "Assume BUFFERS from a frame into the current buffer list.
+
+In interactive use, BUFFERS is determined with a prompt that is
+powered by `completing-read-multiple'.  Multiple candidates can
+be selected, each separated by the `crm-separator' (typically a
+comma).
+
+Also see `beframe-assume-frame-buffers',
+`beframe-unassume-buffers', `beframe-unassume-frame-buffers'."
+  (interactive
+   (list
+    (beframe--buffers-name-to-objects
+     (beframe--buffer-list-prompt-crm
+      (beframe--frame-object
+       (beframe--frame-prompt))))))
+  (beframe--assume buffers))
+
+(defalias 'beframe-add-buffers 'beframe-assume-buffers
+  "Alias of `beframe-assume-buffers' command.")
+
+;;;###autoload
+(defun beframe-assume-buffers-all-frames ()
+  "Like `beframe-assume-buffers' but for the consolidated buffer list (all frames)."
+  (declare (interactive-only t))
+  (interactive)
+  (beframe--assume
+   (beframe--buffers-name-to-objects
+    (beframe--buffer-list-prompt-crm
+     :all-frames))))
+
+;;;###autoload
+(defun beframe-unassume-buffers (buffers)
+  "Unassume BUFFERS from the current frame's buffer list.
+
+In interactive use, BUFFERS is determined with a prompt that is
+powered by `completing-read-multiple'.  Multiple candidates can
+be selected, each separated by the `crm-separator' (typically a
+comma).
+
+Also see `beframe-assume-frame-buffers',
+`beframe-assume-buffers', `beframe-unassume-frame-buffers'."
+  (interactive
+   (list
+    (beframe--buffers-name-to-objects
+     (beframe--buffer-list-prompt-crm))))
+  (beframe--unassume buffers))
+
+(defalias 'beframe-remove-buffers 'beframe-unassume-buffers
+  "Alias of `beframe-unassume-buffers' command.")
+
+;;;###autoload
+(defun beframe-assume-all-buffers-no-prompts ()
+  "Assume the consolidated buffer list (all frames)."
+  (declare (interactive-only t))
+  (interactive)
+  (beframe--assume (beframe--buffer-list-consolidated)))
+
+;;;###autoload
+(defun beframe-unassume-all-buffers-no-prompts ()
+  "Unassume the consolidated buffer list (all frames).
+Keep only the `beframe-global-buffers'."
+  (declare (interactive-only t))
+  (interactive)
+  (beframe--unassume (beframe--buffer-list-consolidated))
+  (beframe--assume (beframe--global-buffers)))
 
 ;;; Minor mode setup
 
@@ -453,7 +514,7 @@ Also see `beframe-assume-frame-buffers'."
                   (when (and beframe-kill-frame-scratch-buffer
                              (null frame))
                     (kill-buffer buf)))))
-    (let* ((frame-bufs (beframe--buffer-list frame))
+    (let* ((frame-bufs (beframe-buffer-list frame))
            (frame-bufs-with-buf (append frame-bufs (list buf))))
       (modify-frame-parameters
        frame
@@ -465,7 +526,7 @@ Also see `beframe-assume-frame-buffers'."
 ;; (defun beframe--rename-scratch-buffer (frame name)
 ;;   "Rename the scratch buffer associated with FRAME according to NAME."
 ;;   (when-let* ((buf (get-buffer (format "*scratch for %s*" frame)))
-;;               ((member (format "*scratch for %s*" frame) (beframe--buffer-list))))
+;;               ((member (format "*scratch for %s*" frame) (beframe-buffer-list))))
 ;;     (with-current-buffer buf
 ;;       (rename-buffer (format "*scratch for %s*" name)))))
 
@@ -529,7 +590,7 @@ Remember that this function doubles as an example for
 (defun beframe--frame-buffer-p (buf &optional frame)
   "Return non-nil if BUF belongs to the current frame.
 Use optional FRAME to test if BUF belongs to it."
-  (memq buf (beframe--buffer-list frame)))
+  (memq buf (beframe-buffer-list frame)))
 
 (defun beframe--frame-predicate (&optional frame)
   "Set FRAME `buffer-predicate' parameter.
@@ -553,5 +614,20 @@ With optional DISABLE remove the advice."
      (beframe-mode
       (advice-add cmd :around #'beframe--with-other-frame)))))
 
+(defun beframe-buffer-sort-visibility (buffers)
+  "Sort the given BUFFERS by visibility.
+Concretely, this means this function will return a sequence that
+first lists hidden, then visible, and then the current buffer."
+  (let* ((current (current-buffer))
+         (bufs (seq-group-by
+                (lambda (buf)
+                  (cond ((eq buf current)                 :current)
+                        ((get-buffer-window buf 'visible) :visible)
+                        (t                                :hidden)))
+                buffers)))
+    (nconc (alist-get :hidden  bufs)
+           (alist-get :visible bufs)
+           (alist-get :current bufs))))
+
 (provide 'beframe)
 ;;; beframe.el ends here
